"""Basic unit tests of parse_command_line()."""
# pylint: disable=invalid-name,import-error
# ruff: noqa: N803

import copy
import unittest
import unittest.mock

from foomuuri import CONFIG as BASE_CONFIG
from foomuuri import INTERNAL as BASE_INTERNAL
from foomuuri import parse_command_line


@unittest.mock.patch('foomuuri.CONFIG_OVERRIDE', new_callable=dict)
@unittest.mock.patch(
    'foomuuri.CONFIG', new_callable=lambda: copy.deepcopy(BASE_CONFIG)
)
@unittest.mock.patch(
    'foomuuri.INTERNAL', new_callable=lambda: copy.deepcopy(BASE_INTERNAL)
)
class TestParseCommandLine(unittest.TestCase):
    """Test parse_command_line()."""

    @unittest.mock.patch('sys.argv', ['foomuuri'])
    def test_no_arguments(self, INTERNAL, *_):
        """Test no arguments."""
        parse_command_line()
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--help'])
    def test_option_help(self, INTERNAL, *_):
        """Test --help option (sets 'help' command)."""
        parse_command_line()
        self.assertEqual(INTERNAL.command, 'help')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--version'])
    def test_option_version(self, INTERNAL, *_):
        """Test --version option (sets 'version' command)."""
        parse_command_line()
        self.assertEqual(INTERNAL.command, 'version')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch(
        'sys.argv',
        ['foomuuri', '--verbose', 'command', 'arg1', '--force', 'arg2'],
    )
    def test_command_and_options(self, INTERNAL, CONFIG, CONFIG_OVERRIDE):
        """Test command with arguments, mixed with options."""
        parse_command_line()
        self.assertEqual(INTERNAL.command, 'command')
        self.assertEqual(INTERNAL.parameters, ['arg1', 'arg2'])
        self.assertEqual(INTERNAL.force, 1)
        self.assertEqual(CONFIG.verbose, 1)
        self.assertEqual(CONFIG_OVERRIDE['verbose'], 1)

    @unittest.mock.patch('sys.argv', ['foomuuri', '--syslog', '--fork'])
    def test_options_syslog_fork(self, INTERNAL, CONFIG, CONFIG_OVERRIDE):
        """Test --syslog and --fork options."""
        parse_command_line()
        self.assertEqual(CONFIG.syslog, 1)
        self.assertEqual(CONFIG.fork, 1)
        self.assertEqual(CONFIG_OVERRIDE['syslog'], 1)
        self.assertEqual(CONFIG_OVERRIDE['fork'], 1)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--verbose'])
    def test_option_verbose(self, INTERNAL, CONFIG, CONFIG_OVERRIDE):
        """Test --verbose option."""
        parse_command_line()
        self.assertEqual(CONFIG.verbose, 1)
        self.assertEqual(CONFIG_OVERRIDE['verbose'], 1)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--verbose', '--verbose'])
    def test_options_verbose_verbose(self, INTERNAL, CONFIG, CONFIG_OVERRIDE):
        """Test --verbose --verbose options."""
        parse_command_line()
        self.assertEqual(CONFIG.verbose, 2)
        self.assertEqual(CONFIG_OVERRIDE['verbose'], 2)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--quiet'])
    def test_option_quiet(self, INTERNAL, CONFIG, CONFIG_OVERRIDE):
        """Test --quiet option."""
        parse_command_line()
        self.assertEqual(CONFIG.verbose, -1)
        self.assertEqual(CONFIG_OVERRIDE['verbose'], -1)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--verbose', '--quiet'])
    def test_options_verbose_quiet(self, INTERNAL, CONFIG, CONFIG_OVERRIDE):
        """Test --verbose --quiet options."""
        parse_command_line()
        self.assertEqual(CONFIG.verbose, 0)
        self.assertEqual(CONFIG_OVERRIDE['verbose'], 0)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--force'])
    def test_option_force(self, INTERNAL, *_):
        """Test --force option."""
        parse_command_line()
        self.assertEqual(INTERNAL.force, 1)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--soft'])
    def test_option_soft(self, INTERNAL, *_):
        """Test --soft option."""
        parse_command_line()
        self.assertEqual(INTERNAL.force, -1)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--set'])
    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_option_set_unknown(self, fail, *_):
        """Test --set option without suffix."""
        self.assertRaises(SystemExit, parse_command_line)
        fail.assert_called_with('Unknown option: --set')

    @unittest.mock.patch('sys.argv', ['foomuuri', '--set='])
    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_option_set_invalid_syntax(self, fail, *_):
        """Test incomplete --set option syntax."""
        self.assertRaises(SystemExit, parse_command_line)
        fail.assert_called_with(
            'Invalid syntax for --set=OPTION=VALUE: --set='
        )

    @unittest.mock.patch('sys.argv', ['foomuuri', '--set=unknown=value'])
    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_option_set_unknown_option(self, fail, *_):
        """Test --set option with unknown foomuuri{} CONFIG option."""
        self.assertRaises(SystemExit, parse_command_line)
        fail.assert_called_with('Unknown foomuuri{} option: unknown')

    @unittest.mock.patch('sys.argv', ['foomuuri', '--set=priority_offset=yes'])
    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_option_set_invalid_value(self, fail, *_):
        """Test --set option with invalid value."""
        self.assertRaises(SystemExit, parse_command_line)
        fail.assert_called_with(
            'Invalid foomuuri{} option priority_offset value: yes'
        )

    @unittest.mock.patch(
        'sys.argv',
        ['foomuuri', '--set=priority_offset=42', '--set=set_size=1'],
    )
    def test_options_set_valid(self, INTERNAL, CONFIG, _):
        """Test --set options with valid foomuuri{} CONFIG option/value."""
        parse_command_line()
        self.assertEqual(CONFIG.priority_offset, 42)
        self.assertEqual(CONFIG.set_size, 1)
        self.assertEqual(INTERNAL.command, '')
        self.assertEqual(INTERNAL.parameters, [])

    @unittest.mock.patch('sys.argv', ['foomuuri', '--unknown'])
    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_unknown_option(self, fail, *_):
        """Test unknown option."""
        self.assertRaises(SystemExit, parse_command_line)
        fail.assert_called_with('Unknown option: --unknown')

    @unittest.mock.patch('sys.argv', ['foomuuri', 'command', '-h', '-v'])
    def test_option_short(self, INTERNAL, *_):
        """Test short options. They are parsed as parameters."""
        parse_command_line()
        self.assertEqual(INTERNAL.command, 'command')
        self.assertEqual(INTERNAL.parameters, ['-h', '-v'])

    @unittest.mock.patch('sys.argv', ['foomuuri', 'command', 'a', '-', 'b'])
    def test_option_hyphen(self, INTERNAL, *_):
        """Test hyphen. It is parsed as parameter."""
        parse_command_line()
        self.assertEqual(INTERNAL.command, 'command')
        self.assertEqual(INTERNAL.parameters, ['a', '-', 'b'])



"""Basic unit tests of Converters."""
# pylint: disable=import-error

import unittest

from foomuuri import Converters


class TestConverters(unittest.TestCase):
    """Test converter helpers."""

    def test_str_yes_no_to_bool(self):
        """Test str_yes_no_to_bool."""
        self.assertTrue(Converters.str_yes_no_to_bool('yes'))
        self.assertFalse(Converters.str_yes_no_to_bool('no'))
        self.assertRaises(ValueError, Converters.str_yes_no_to_bool, '')
        self.assertRaises(ValueError, Converters.str_yes_no_to_bool, 'invalid')



"""Basic unit tests of print_table()."""
# pylint: disable=invalid-name,import-error

import io
import unittest
import unittest.mock

from foomuuri import print_table


@unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
class TestPrintTable(unittest.TestCase):
    """Test print_table()."""

    def setUp(self):
        """Fill in common fixture values."""
        self.data_multiple_rows = [
            (
                'row 1 column 1',
                'row 1 colum 2',
                'row 1 colu 3',
            ),
            (
                'row 2 colu 1',
                'row 2 colum 2',
                'row 2 column 3',
            ),
        ]

        self.data_no_rows = []

        self.header = (
            'Header',
            'Long Elaborate Header',
            'Medium Header',
        )

    def test_invalid_align_values(self, _):
        """Test invalid align values."""
        invalid_align = [
            '',  # no alignment
            'x',  # invalid align value
            '<<',  # too few values for 3 columns
            '<<<<',  # too much values for 3 columns
        ]

        for align in invalid_align:
            with self.subTest(align):
                self.assertRaises(
                    ValueError,
                    print_table,
                    self.data_multiple_rows,
                    align=align,
                )

    def test_print_table_all_args(self, stdout: io.StringIO):
        """Test table with all supported arguments."""
        print_table(
            data=[self.header] + self.data_multiple_rows,
            header=True,
            align='<>^',
            sep='    ',
        )

        self.assertEqual(
            stdout.getvalue(),
            """Header            Long Elaborate Header    Medium Header
--------------    ---------------------    --------------
row 1 column 1            row 1 colum 2     row 1 colu 3
row 2 colu 1              row 2 colum 2    row 2 column 3
""",
        )

    def test_print_table_with_rows_no_headers(self, stdout: io.StringIO):
        """Test table without headers, rows only."""
        print_table(data=self.data_multiple_rows)

        self.assertEqual(
            stdout.getvalue(),
            """row 1 column 1   row 1 colum 2   row 1 colu 3
row 2 colu 1     row 2 colum 2   row 2 column 3
""",
        )

    def test_print_table_no_rows(self, stdout: io.StringIO):
        """Test table without headers and rows."""
        print_table(data=self.data_no_rows)

        self.assertEqual(
            stdout.getvalue(),
            '',
        )

    def test_print_table_with_headers_no_rows(self, stdout: io.StringIO):
        """Test table with headers, without rows."""
        print_table(data=[self.header] + self.data_no_rows, header=True)
        self.assertEqual(
            stdout.getvalue(),
            '',
        )

    def test_print_table_default_align(self, stdout: io.StringIO):
        """Test table with headers, and default (left) align."""
        print_table(data=[self.header] + self.data_multiple_rows, header=True)

        self.assertEqual(
            stdout.getvalue(),
            """Header           Long Elaborate Header   Medium Header
--------------   ---------------------   --------------
row 1 column 1   row 1 colum 2           row 1 colu 3
row 2 colu 1     row 2 colum 2           row 2 column 3
""",
        )



"""Basic unit tests of TypedConfig."""
# pylint: disable=import-error

import pathlib
import unittest
from dataclasses import dataclass, field

from foomuuri import TypedConfig


class TestTypedConfig(unittest.TestCase):
    """Setters / getters / type conversion tests."""

    def set(self, name, value):
        """Set attribute and return value helper."""
        self.config[name] = value
        return self.config[name]

    def set_from_str(self, name, value):
        """Set attribute from string and return value helper."""
        self.config.set_from_str(name, value)
        return self.config[name]

    def append_from_str(self, name, value):
        """Append attribute value from string and return value helper."""
        self.config.append_from_str(name, value)
        return self.config[name]

    def setUp(self):
        """Setup tests."""

        @dataclass
        class TestConfig(TypedConfig):
            """Test class."""

            # pylint: disable=too-many-instance-attributes
            initialized_str: str = 'init_value1'
            uninitialized_str: str = field(init=False)
            initialized_untyped = 'init_value2'  # noqa: RUF100,RUF045
            type_conversion_float: float = 0.0
            type_conversion_int: int = 0
            type_conversion_list: list = field(default_factory=list)
            type_conversion_posixpath: pathlib.PosixPath = field(init=False)
            type_conversion_set: set = field(default_factory=set)
            validation: str = field(
                init=False, metadata={'validate': lambda v: v}
            )
            conversion: list = field(
                init=False, metadata={'convert_str': lambda v: v.split()}
            )

        self.config = TestConfig()

    def test_accessor_style(self):
        """Test dict/dot style accessor/setter."""
        self.config.initialized_str = 'new_value1'
        self.assertEqual(self.config.initialized_str, 'new_value1')
        self.assertRaises(
            AttributeError, lambda: self.config.unknown_attribute
        )

        self.config['initialized_str'] = 'new_value2'
        self.assertEqual(self.config['initialized_str'], 'new_value2')
        self.assertRaises(
            AttributeError, lambda: self.config['unknown_attribute']
        )

        self.assertRaises(AttributeError, self.set, 'unknown_attribute', '123')

    def test_initialized_typed(self):
        """Test assigning values to initialized typed attributes."""
        self.assertRaises(TypeError, self.set, 'initialized_str', 123)
        self.assertEqual(self.set('initialized_str', 'value'), 'value')

    def test_uninitialized_typed(self):
        """Test assigning values to uninialized typed attributes."""
        self.assertRaises(
            AttributeError, lambda: self.config.uninitialized_str
        )
        self.assertRaises(TypeError, self.set, 'uninitialized_str', 123)
        self.assertEqual(self.set('uninitialized_str', 'value'), 'value')

    def test_initialized_untyped(self):
        """Test accessing listed untyped attributes."""
        self.assertRaises(
            AttributeError, lambda: self.config.uninitialized_untyped
        )

    def test_set_from_str(self):
        """Test attribute type conversion from str (set_from_str)."""
        # Invalid value type: only str supported
        self.assertRaises(
            TypeError, self.config.set_from_str, 'type_conversion_int', 123
        )
        # Supported conversion, from str to int
        self.assertEqual(self.set_from_str('type_conversion_int', '123'), 123)
        # Supported conversion, from str to pathlib.PosixPath
        self.assertEqual(
            self.set_from_str('type_conversion_posixpath', '/path/'),
            pathlib.PosixPath('/path/'),
        )
        # Supported conversion, from str to list
        self.assertEqual(
            self.set_from_str('type_conversion_list', '1 2 2 3 4'),
            ['1', '2', '2', '3', '4'],
        )
        # Supported conversion, from str to set
        self.assertEqual(
            self.set_from_str('type_conversion_set', '1 2 2 3 4'),
            {'1', '2', '3', '4'},
        )
        # Assignment, str to str
        self.assertEqual(self.set_from_str('initialized_str', '123'), '123')
        # Assignment, list converted from str
        self.assertEqual(self.set_from_str('conversion', '1'), ['1'])
        # Unsupported conversion, from str to float
        self.assertRaises(
            TypeError,
            self.config.set_from_str,
            'type_conversion_float',
            '42.0',
        )
        # Assign non-existing attribute
        self.assertRaises(
            AttributeError, self.config.set_from_str, 'unknownattribute', '123'
        )

    def test_append_from_str(self):
        """Test attribute value append from str (append_from_str)."""
        # Invalid value type: only str supported
        self.assertRaises(
            TypeError, self.config.append_from_str, 'initialized_str', 123
        )
        # Appending, str to str
        self.assertEqual(self.set_from_str('initialized_str', '1'), '1')
        self.assertEqual(self.append_from_str('initialized_str', '2'), '1 2')
        # Appending, list converted from str to list
        self.assertEqual(self.set_from_str('conversion', '1'), ['1'])
        self.assertEqual(self.append_from_str('conversion', '2'), ['1', '2'])
        # Appending, str to list
        self.assertEqual(
            self.set_from_str('type_conversion_list', '1 2'),
            ['1', '2'],
        )
        self.assertEqual(
            self.append_from_str('type_conversion_list', '2 3'),
            ['1', '2', '2', '3'],
        )
        # Appending, str to set
        self.assertEqual(
            self.set_from_str('type_conversion_set', '1 2'),
            {'1', '2'},
        )
        self.assertEqual(
            self.append_from_str('type_conversion_set', '2 3'),
            {'1', '2', '3'},
        )
        # Unsupported append, from str to int
        self.assertRaises(
            TypeError, self.append_from_str, 'type_conversion_int', '1'
        )
        # Append non-existing attribute
        self.assertRaises(
            AttributeError, self.append_from_str, 'unknownattribute', '123'
        )

    def test_validation(self):
        """Test attribute validation helper call."""
        self.assertRaises(ValueError, self.set, 'validation', '')
        self.assertEqual(self.set('validation', 'test'), 'test')

    def test_iter(self):
        """Test __iter__."""
        self.assertEqual(
            sorted(self.config),
            [
                'conversion',
                'initialized_str',
                'type_conversion_float',
                'type_conversion_int',
                'type_conversion_list',
                'type_conversion_posixpath',
                'type_conversion_set',
                'uninitialized_str',
                'validation',
            ],
        )

    def test_get(self):
        """Test get."""
        self.assertEqual(self.config.get('initialized_str'), 'init_value1')
        self.assertEqual(self.config.get('unknown_attr'), None)



"""Basic unit tests of parse_command_line()."""
# pylint: disable=invalid-name,import-error
# ruff: noqa: N803

import contextlib
import copy
import unittest
import unittest.mock

from foomuuri import CONFIG as BASE_CONFIG
from foomuuri import INTERNAL as BASE_INTERNAL
from foomuuri import minimal_config


@unittest.mock.patch(
    'foomuuri.INTERNAL', new_callable=lambda: copy.deepcopy(BASE_INTERNAL)
)
@unittest.mock.patch(
    'foomuuri.CONFIG', new_callable=lambda: copy.deepcopy(BASE_CONFIG)
)
class TestParseConfigFoomuuri(unittest.TestCase):
    """Test parse_config_foomuuri()."""

    @staticmethod
    @contextlib.contextmanager
    def mock_config_foomuuri(option, value):
        """Return context manager of foomuuri.find_config_files() patch.

        foomuuri.find_config_files() is patched to return generated
        foomuuri configuration file, containing foomuuri{} section
        with single option/value.
        """
        # MagicMock of pathlib.PosixPAth returning desired foomuuri config
        config_file = unittest.mock.MagicMock()
        config_file.read_text.return_value = f"""
            foomuuri {{
                {option} {value}
            }}
        """
        # foomuuri.find_config_files() is called by foomuuri.read_config()
        # to load configuration files.
        # read_config() tries loading share_config first, and etc_config then.
        # We only care about etc_config, therefore side_effect list passed
        # to foomuuri.find_config_files mock patch has two elemeents,
        # [] - empty share_config, and [config_file] - generated config.
        with unittest.mock.patch(
            'foomuuri.find_config_files', side_effect=[[], [config_file]]
        ):
            yield config_file

    @unittest.mock.patch('foomuuri.warning')
    def test_try_reload_timeout(self, warning, CONFIG, *_):
        """Test try-reload_timeout backwards compatible option."""
        with self.mock_config_foomuuri(
            'try-reload_timeout',
            69,
        ) as config_file:
            _ = minimal_config()
            self.assertEqual(CONFIG.try_reload_timeout, 69)
            warning.assert_called_once_with(
                f'File {config_file} line 3: foomuuri{{}} option '
                f'"try-reload_timeout" is obsolete, use '
                f'"try_reload_timeout" instead'
            )

    def test_append_value(self, CONFIG, *_):
        """Test appending of option value."""
        with self.mock_config_foomuuri('log_level', '+ append'):
            _ = minimal_config()
            self.assertEqual(CONFIG.log_level, 'level info flags skuid append')

    def test_unknown_option(self, *_):
        """Test assigning unknown option."""
        with (
            self.mock_config_foomuuri(
                'unknown_option', 'value'
            ) as config_file,
            unittest.mock.patch(
                'foomuuri.fail', side_effect=SystemExit
            ) as fail,
        ):
            self.assertRaises(SystemExit, minimal_config)
            fail.assert_called_once_with(
                f'File {config_file} line 3: Unknown foomuuri{{}}'
                ' option: unknown_option value'
            )

    def test_invalid_value(self, *_):
        """Test assigning invalid value."""
        with (
            self.mock_config_foomuuri(
                'priority_offset', 'not_an_int'
            ) as config_file,
            unittest.mock.patch(
                'foomuuri.fail', side_effect=SystemExit
            ) as fail,
        ):
            self.assertRaises(SystemExit, minimal_config)
            fail.assert_called_once_with(
                f'File {config_file} line 3: Invalid foomuuri{{}}'
                ' option value: priority_offset not_an_int'
            )

    @unittest.mock.patch('foomuuri.CONFIG_OVERRIDE', new_callable=dict)
    def test_config_override(self, CONFIG_OVERRIDE, CONFIG, *_):
        """Test overriding of CONFIG from CONFIG_OVERRRIDE."""

        def test_override(option, override, config):
            with self.mock_config_foomuuri(option, '0'):
                CONFIG_OVERRIDE[option] = override
                _ = minimal_config()
                self.assertEqual(CONFIG[option], config)

        # Overriding with correct value type (str)
        test_override(option='priority_offset', override='69', config=69)
        # Overriding with invalid value type (int).
        test_override(option='priority_offset', override=69, config=69)

    def test_postprocess_priority_offset(self, _, INTERNAL):
        """Test postprocessing of priority_offset option."""

        def test_priority_offset(config, internal):
            with self.mock_config_foomuuri('priority_offset', config):
                _ = minimal_config()
                self.assertEqual(INTERNAL.priority_offset, internal)

        test_priority_offset(config=-1, internal=' - 1')
        test_priority_offset(config=0, internal='')
        test_priority_offset(config=1, internal=' + 1')

    def test_postprocess_log_rates(self, CONFIG, *_):
        """Test postprocessing of log_* rates options."""
        with self.mock_config_foomuuri('log_rate', '1/second burst 3'):
            _ = minimal_config()
            self.assertTrue(CONFIG.log_rate.endswith(' packets'))



"""Basic unit tests of IPList classes."""
# pylint: disable=invalid-name,import-error

import collections.abc
import unittest

from foomuuri import IPList, IPLists


class TestIPLists(unittest.TestCase):
    """Test IPLists."""

    def setUp(self):
        """Prepare test fixture."""
        self.iplists = IPLists()

        self.sources_foo = ['https://foob.ar', 'https://foob.az']
        self.iplists['@foo'] = IPList(sources=self.sources_foo)

        self.sources_bar = ['https://arbo.of', 'https://azbp.of']
        self.iplists['@bar'] = IPList(sources=self.sources_bar)
        self.iplists['@bar'].options.start = False

    def test_index_getter(self):
        """Test index getter by iplist name."""
        self.assertEqual(self.iplists['@foo'].sources, self.sources_foo)
        self.assertEqual(self.iplists['@bar'].sources, self.sources_bar)

    def test_iter_names(self):
        """Test iterator over iplist names."""
        self.assertEqual(list(self.iplists), ['@foo', '@bar'])

    def test_items(self):
        """Test iterable over iplist entries (name, value tuples)."""
        self.assertIsInstance(self.iplists.items(), collections.abc.Iterable)
        self.assertListEqual(
            list(self.iplists.items()),
            [
                ('@foo', self.iplists['@foo']),
                ('@bar', self.iplists['@bar']),
            ],
        )

    def test_values(self):
        """Test iterable over iplist entries values."""
        self.assertIsInstance(self.iplists.values(), collections.abc.Iterable)
        self.assertListEqual(
            list(self.iplists.values()),
            [
                self.iplists['@foo'],
                self.iplists['@bar'],
            ],
        )

    def test_filter_names_by(self):
        """Test set of iplist names with specific iplist option value."""
        self.assertIsInstance(self.iplists.filter_names_by('start', True), set)
        self.assertEqual(self.iplists.filter_names_by('start', True), {'@foo'})



"""Basic unit tests of join_args()."""
# pylint: disable=import-error

import pathlib
import unittest

from foomuuri import join_args


class TestJoinArgs(unittest.TestCase):
    """Test join_args()."""

    def test_valid(self):
        """Test valid outcomes."""
        self.assertEqual(join_args([]), '')
        self.assertEqual(join_args(['ab', 'cd']), 'ab cd')
        self.assertEqual(join_args(['ab', pathlib.PosixPath('cd')]), 'ab cd')

    def test_invalid(self):
        """Test invalid outcomes."""
        self.assertRaises(TypeError, join_args, 'ab')



"""Hack to import foomuuri without .py extension and proper package tree."""

import sys
from importlib.abc import Loader
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader

if (  # noqa: RUF100,RUF067
    (
        spec := spec_from_loader(
            'foomuuri', SourceFileLoader('foomuuri', './src/foomuuri')
        )
    )
    and (foomuuri := module_from_spec(spec))
    and isinstance(spec.loader, Loader)
):
    spec.loader.exec_module(foomuuri)
    sys.modules['foomuuri'] = foomuuri
else:
    raise SystemError('Failed to add foomuuri to sys.modules')



"""Basic unit tests of parse_duration() and seconds_to_duration()."""
# pylint: disable=import-error

import unittest

from foomuuri import parse_duration, seconds_to_duration


class TestParseDuration(unittest.TestCase):
    """Test parse_duration()."""

    def test_valid_duration(self):
        """Test valid outcomes."""
        self.assertEqual(parse_duration(' 0s ', fallback=None), 0)
        self.assertEqual(parse_duration(' 1s ', fallback=0), 1)
        self.assertEqual(parse_duration(' 1h 1m ', fallback=0), 3660)
        self.assertEqual(parse_duration(' 1w1d1h1m1s ', fallback=0), 694861)
        self.assertEqual(parse_duration(' 1w_1d_1h_1m_1s', fallback=0), 694861)

    def test_invalid_duration(self):
        """Test invalid outcomes."""
        self.assertIsNone(parse_duration('', fallback=None))
        self.assertIsNone(parse_duration('1day', fallback=None))
        self.assertIsNone(parse_duration('1y', fallback=None))
        self.assertIsNone(parse_duration('s', fallback=None))
        self.assertIsNone(parse_duration('1s1h', fallback=None))


class TestSecondsToDuration(unittest.TestCase):
    """Test seconds_to_duration()."""

    def test_valid_seconds(self):
        """Test valid outcomes."""
        self.assertEqual(seconds_to_duration(0), '0s')
        self.assertEqual(seconds_to_duration(1), '1s')
        self.assertEqual(seconds_to_duration(42), '42s')
        self.assertEqual(seconds_to_duration(61), '1m01s')
        self.assertEqual(seconds_to_duration(3600), '1h00m00s')
        self.assertEqual(seconds_to_duration(99999), '1d03h46m39s')
        self.assertEqual(seconds_to_duration(999999), '11d13h46m39s')
        self.assertEqual(seconds_to_duration(9999999), '115d17h46m39s')



"""Basic unit tests of Validators."""
# pylint: disable=import-error

import unittest

from foomuuri import Validators


class TestValidators(unittest.TestCase):
    """Test assert helpers."""

    def test_assert_str_word(self):
        """Test invalid and valid outcomes."""
        self.assertFalse(Validators.str_zone_name(''))
        self.assertFalse(Validators.str_zone_name('word word'))
        self.assertFalse(Validators.str_zone_name('3word'))
        self.assertFalse(Validators.str_zone_name('w#ord'))
        self.assertFalse(Validators.str_zone_name('_./_'))
        self.assertTrue(Validators.str_zone_name('wo_123_rd'))
        self.assertTrue(Validators.str_zone_name('w./_'))

        self.assertFalse(Validators.str_iplist_name(''))
        self.assertFalse(Validators.str_iplist_name('@'))
        self.assertFalse(Validators.str_iplist_name('@w w'))
        self.assertFalse(Validators.str_iplist_name('@kääk'))
        self.assertFalse(Validators.str_iplist_name('@3word'))
        self.assertTrue(Validators.str_iplist_name('@word'))
        self.assertTrue(Validators.str_iplist_name('@_'))
        self.assertTrue(Validators.str_iplist_name('@_./-'))

        self.assertFalse(Validators.str_interface_name(''))
        self.assertFalse(Validators.str_interface_name('word word'))
        self.assertFalse(Validators.str_interface_name('word\nword'))
        self.assertFalse(Validators.str_interface_name('word/word'))
        self.assertFalse(Validators.str_interface_name('word"word'))
        self.assertFalse(Validators.str_interface_name('word\\word'))
        self.assertTrue(Validators.str_interface_name('eth0'))
        self.assertTrue(Validators.str_interface_name(':'))
        self.assertTrue(Validators.str_interface_name('@'))
        self.assertTrue(Validators.str_interface_name('kääk'))

        self.assertFalse(Validators.str_words(''))
        self.assertTrue(Validators.str_words('word'))
        self.assertTrue(Validators.str_words('word word'))

        self.assertFalse(Validators.has_elements([]))
        self.assertTrue(Validators.has_elements(['a']))
        self.assertTrue(Validators.has_elements(['a', 'b']))
        self.assertFalse(Validators.has_elements({}))
        self.assertTrue(Validators.has_elements({'a'}))
        self.assertTrue(Validators.has_elements({'a', 'b'}))

        self.assertFalse(Validators.int_positive_or_zero(-1))
        self.assertTrue(Validators.int_positive_or_zero(0))
        self.assertTrue(Validators.int_positive_or_zero(1))

        self.assertFalse(Validators.int_positive(-1))
        self.assertTrue(Validators.int_positive(1))

        self.assertFalse(Validators.str_yes_no(''))
        self.assertFalse(Validators.str_yes_no('maybe'))
        self.assertFalse(Validators.str_yes_no('YES'))
        self.assertTrue(Validators.str_yes_no('yes'))
        self.assertTrue(Validators.str_yes_no('no'))



"""Basic unit tests of parse_config_iplist()."""
# pylint: disable=invalid-name,import-error
# ruff: noqa: N803

import contextlib
import copy
import unittest
import unittest.mock

from foomuuri import CONFIG as BASE_CONFIG
from foomuuri import INTERNAL as BASE_INTERNAL
from foomuuri import (
    IPListOptions,
    minimal_config,
    parse_config_iplist,
    parse_duration,
)


@unittest.mock.patch(
    'foomuuri.INTERNAL', new_callable=lambda: copy.deepcopy(BASE_INTERNAL)
)
@unittest.mock.patch(
    'foomuuri.CONFIG', new_callable=lambda: copy.deepcopy(BASE_CONFIG)
)
class TestIplistParseConfig(unittest.TestCase):
    """Test parse_config_iplist()."""

    @staticmethod
    @contextlib.contextmanager
    def mock_config_foomuuri(section: str = 'iplist', content: str = ''):
        """Return context manager of foomuuri.find_config_files() patch.

        foomuuri.find_config_files() is patched to return generated
        foomuuri configuration file, containing arbitrary section {}
        with specified content.
        """
        # consult tests/test_parse_config_foomuuri.py for mock explanation
        config_file = unittest.mock.MagicMock()
        config_file.read_text.return_value = f"""
            {section} {{
                {content}
            }}
        """
        with unittest.mock.patch(
            'foomuuri.find_config_files', side_effect=[[], [config_file]]
        ):
            yield config_file

    @unittest.mock.patch('foomuuri.warning')
    def test_obsolete_syntax(self, warning, *_):
        """Test obsolete resolve{} section and timeout/refresh options."""
        with self.mock_config_foomuuri(
            section='resolve',
            content="""
            timeout 1d 5m
            refresh 0d 30m
            @foo https://foob.ar/
        """,
        ) as config_file:
            iplists = parse_config_iplist(minimal_config())
            options = iplists['@foo'].options
            self.assertEqual(options.dns_timeout, parse_duration('1d 5m', 0))
            self.assertEqual(options.url_timeout, parse_duration('1d 5m', 0))
            self.assertEqual(options.dns_refresh, parse_duration('0d 30m', 0))
            self.assertEqual(options.url_refresh, parse_duration('0d 30m', 0))
            self.assertEqual(iplists['@foo'].sources, ['https://foob.ar/'])
            warning.assert_has_calls(
                [
                    unittest.mock.call(
                        'Section "resolve" is obsolete, use "iplist" instead',
                    ),
                    unittest.mock.call(
                        f'File {config_file} line 4: Iplist "timeout" is '
                        'obsolete, use "dns_timeout" or "url_timeout" instead',
                    ),
                    unittest.mock.call(
                        f'File {config_file} line 5: Iplist "refresh" is '
                        'obsolete, use "dns_refresh" or "url_refresh" instead',
                    ),
                ]
            )

    @unittest.mock.patch('foomuuri.warning')
    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_obsolete_syntax_fail(self, fail, warning, *_):
        """Test iplist{} section with invalid obsolete timeout option."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            timeout  # value is missing
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())
            warning.assert_called_once_with(
                f'File {config_file} line 4: Iplist "timeout" is obsolete, '
                'use "dns_timeout" or "url_timeout" instead',
            )
            fail.assert_called_once_with(
                f'File {config_file} line 4: Invalid iplist{{}} option '
                'value: dns_timeout='
            )

    def test_empty_iplist_definition(self, *_):
        """Test empty iplist definition."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo
            @foo -
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            self.assertEqual(iplists['@foo'].sources, [])

    def test_valid_default_options(self, *_):
        """Test valid default options, space and equal-sign styles."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            url_max_size 1024
            url_timeout=69d
            url_refresh=42d
            dns_timeout 1d 2h  # space-separated option value is
                               # supported by parse_duration()
            @foo https://foob.ar/
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            options = iplists['@foo'].options
            self.assertEqual(options.url_timeout, parse_duration('69d', 0))
            self.assertEqual(options.url_refresh, parse_duration('42d', 0))
            self.assertEqual(options.dns_timeout, parse_duration('1d 2h', 0))
            self.assertEqual(options.url_max_size, 1024)

    def test_single_line_multiple_valid_default_options(self, *_):
        """Test mulitple valid default options on single line."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            url_timeout=69d url_refresh=42d
            @foo https://foob.ar/
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            options = iplists['@foo'].options
            self.assertEqual(options.url_timeout, parse_duration('69d', 0))
            self.assertEqual(options.url_refresh, parse_duration('42d', 0))

    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_first_invalid_default_option(self, fail, *_):
        """Test first invalid default option on single line."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            invalid url_timeout=2d url_refresh=1d
            @foo https://foob.ar/
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())

            fail.assert_called_once_with(
                f'File {config_file} line 4: Invalid iplist{{}} option name: '
                'invalid=url_timeout=2d url_refresh=1d'
            )

    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_second_invalid_default_option(self, fail, *_):
        """Test second invalid default option on single line."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            url_timeout=2d invalid url_refresh=1d
            @foo https://foob.ar/
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())

            fail.assert_called_once_with(
                f'File {config_file} line 4: Unsupported syntax: '
                'url_timeout=2d invalid url_refresh=1d'
            )

    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_invalid_space_style_default_option_name(self, fail, *_):
        """Test invalid space-style default option name."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            url_min_size 1024
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())

            fail.assert_called_once_with(
                f'File {config_file} line 4: Invalid iplist{{}} option name: '
                'url_min_size=1024'
            )

    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_invalid_space_style_default_option_value(self, fail, *_):
        """Test invalid space-style default option value."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            url_max_size abcd
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())

            fail.assert_called_once_with(
                f'File {config_file} line 4: Invalid iplist{{}} option value: '
                'url_max_size=abcd'
            )

    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_invalid_default_option_name(self, fail, *_):
        """Test invalid default option name."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            invalid_option=whatever_value
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())

            fail.assert_called_once_with(
                f'File {config_file} line 4: Invalid iplist{{}} option name: '
                'invalid_option=whatever_value'
            )

    @unittest.mock.patch('foomuuri.fail', side_effect=SystemExit)
    def test_invalid_default_option_value(self, fail, *_):
        """Test invalid default option value."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            url_max_size=abcd
        """,
        ) as config_file:
            with self.assertRaises(SystemExit):
                _ = parse_config_iplist(minimal_config())

            fail.assert_called_once_with(
                f'File {config_file} line 4: Invalid iplist{{}} option value: '
                'url_max_size=abcd'
            )

    def test_start_merge_url_size_options(self, *_):
        """Test start, merge, url_size options."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            -start -merge url_max_size=1024
            @foo https://foob.ar/
            @bar https://b.ar/ start=no merge=no
            @baz https://b.az/ start=yes merge=yes url_max_size=2048
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            default_options = IPListOptions()
            # Testing applcation as default options
            # Asserting applied options differ from default values
            self.assertNotEqual(
                iplists['@foo'].options.start, default_options.start
            )
            self.assertNotEqual(
                iplists['@foo'].options.merge, default_options.merge
            )
            self.assertNotEqual(
                iplists['@foo'].options.url_max_size,
                default_options.url_max_size,
            )
            # Asserting applied option values are exactly as set
            self.assertFalse(iplists['@foo'].options.start)
            self.assertFalse(iplists['@foo'].options.merge)
            self.assertFalse(iplists['@foo'].options.dynamic)
            self.assertEqual(iplists['@foo'].options.url_max_size, 1024)
            # Alternative start/merge syntax (false)
            self.assertFalse(iplists['@bar'].options.start)
            self.assertFalse(iplists['@bar'].options.merge)
            # Alternative start/merge syntax (true)
            self.assertTrue(iplists['@baz'].options.start)
            self.assertTrue(iplists['@baz'].options.merge)
            # Iplist url_max_size overrides default option
            self.assertTrue(iplists['@baz'].options.url_max_size, 2048)

    @unittest.mock.patch('foomuuri.warning')
    def test_iplist_sources_overwrite(self, warning, *_):
        """Test overwriting of iplist sources."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo https://foob.ar/
            @foo https://rabo.of/
        """,
        ) as config_file:
            iplists = parse_config_iplist(minimal_config())
            self.assertEqual(
                iplists['@foo'].sources,
                ['https://rabo.of/'],
            )
            warning.assert_called_with(
                f'File {config_file} line 5: Overwriting iplist "@foo" '
                'with value "https://rabo.of/"'
            )

    def test_iplist_sources_append(self, *_):
        """Test appending of iplist sources."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo https://foob.ar/
            @foo + https://rabo.of/

            @bar + https://foo.bar/

            @baz +
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            self.assertEqual(
                iplists['@foo'].sources,
                ['https://foob.ar/', 'https://rabo.of/'],
            )
            self.assertEqual(iplists['@bar'].sources, ['https://foo.bar/'])
            self.assertEqual(iplists['@baz'].sources, [])

    def test_intermixing_iplist_sources_options(self, *_):
        """Test itermixing of iplist sources and options."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo start=no https://foob.ar/ merge=no https://rabo.of/
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            default_options = IPListOptions()
            # Asserting applied options differ from default values
            self.assertNotEqual(
                iplists['@foo'].options.start, default_options.start
            )
            self.assertNotEqual(
                iplists['@foo'].options.merge, default_options.merge
            )
            # Asserting appled options values are exactly as set
            self.assertFalse(iplists['@foo'].options.start)
            self.assertFalse(iplists['@foo'].options.merge)
            self.assertEqual(
                iplists['@foo'].sources,
                ['https://foob.ar/', 'https://rabo.of/'],
            )

    def test_last_option_priority_wins(self, *_):
        """Test last option defined for same ipset definition wins."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo url_timeout=42d
            @foo url_timeout=69d
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            self.assertEqual(
                iplists['@foo'].options.url_timeout, parse_duration('69d', 0)
            )

    @unittest.mock.patch('foomuuri.verbose')
    def test_iplist_name_and_sources_vebose_output(self, verbose, CONFIG, *_):
        """Test output of iplists name and sources when verbose >= 2."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo https://foob.ar/
            @bar https://foo.bar/
        """,
        ):
            CONFIG.verbose = 2
            _ = parse_config_iplist(minimal_config())
            verbose.assert_has_calls(
                [
                    unittest.mock.call('@foo       https://foob.ar/'),
                    unittest.mock.call('@bar       https://foo.bar/'),
                ]
            )

    def test_iplist_dynamic(self, *_):
        """Test dynamic flag in iplist."""
        with self.mock_config_foomuuri(
            section='iplist',
            content="""
            @foo dynamic=yes
            @bar dynamic=yes element_timeout=1m30s
        """,
        ):
            iplists = parse_config_iplist(minimal_config())
            self.assertTrue(iplists['@foo'].options.dynamic)
            self.assertTrue(iplists['@bar'].options.dynamic)
            self.assertEqual(
                iplists['@foo'].options.element_timeout,
                parse_duration('0s', -1),
            )
            self.assertEqual(
                iplists['@bar'].options.element_timeout,
                parse_duration('1m30s', -1),
            )

    def test_internal_iplist_missing_ok_init(self, _, INTERNAL, *__):
        """Test INTERNAL.iplist_missing_ok is initialized."""
        with self.mock_config_foomuuri():
            self.assertRaises(
                AttributeError, lambda: INTERNAL.iplist_missing_ok
            )
            _ = parse_config_iplist(minimal_config())
            self.assertEqual(INTERNAL.iplist_missing_ok, set())



"""Basic unit tests for test_ipv*_address() functions."""
# pylint: disable=import-error

import unittest

from foomuuri import is_ip_address, is_ipv4_address, is_ipv6_address


class TestIsIPv4Address(unittest.TestCase):
    """Basic unit tests for test_ipv4_address()."""

    def test_ipv4_address(self):
        """Test for IPv4 address."""
        self.assertFalse(is_ipv4_address('', allow_network=False))
        self.assertFalse(is_ipv4_address('::', allow_network=False))
        self.assertFalse(is_ipv4_address('127.0.0.0/8', allow_network=False))
        self.assertTrue(is_ipv4_address('127.0.0.1', allow_network=False))

    def test_ipv4_network(self):
        """Test for IPv4 network."""
        self.assertFalse(is_ipv4_address('', allow_network=True))
        self.assertFalse(is_ipv4_address('::/64', allow_network=True))
        self.assertTrue(is_ipv4_address('127.0.0.0/8', allow_network=True))
        self.assertTrue(is_ipv4_address('127.0.0.1/8', allow_network=True))

    def test_ipv4_range(self):
        """Test for IPv4 range."""
        self.assertFalse(is_ipv4_address('-'))
        self.assertFalse(is_ipv4_address('::-::'))
        self.assertFalse(is_ipv4_address('127.0.0.0/8-127.0.0.0/8'))
        self.assertFalse(is_ipv4_address('127.0.0.2-127.0.0.1'))
        self.assertTrue(is_ipv4_address('127.0.0.1-127.0.0.1'))
        self.assertTrue(is_ipv4_address('127.0.0.1-127.0.0.2'))


class TestIsIPv6Address(unittest.TestCase):
    """Basic unit tests for test_ipv6_address()."""

    def test_ipv6_address(self):
        """Test for IPv6 address."""
        self.assertFalse(is_ipv6_address('', allow_network=False))
        self.assertFalse(is_ipv6_address('127.0.0.1', allow_network=False))
        self.assertFalse(is_ipv6_address('[::]/64', allow_network=False))
        self.assertTrue(is_ipv6_address('::', allow_network=False))
        self.assertTrue(is_ipv6_address('[::]', allow_network=False))

    def test_ipv6_network(self):
        """Test for IPv6 network."""
        self.assertFalse(is_ipv6_address('', allow_network=True))
        self.assertFalse(is_ipv6_address('127.0.0.0/8', allow_network=True))
        self.assertFalse(is_ipv6_address('::/129', allow_network=True))
        self.assertFalse(is_ipv6_address('[::]/-64', allow_network=True))
        self.assertTrue(is_ipv6_address('::/64', allow_network=True))
        self.assertTrue(is_ipv6_address('::/-64', allow_network=True))
        self.assertTrue(is_ipv6_address('[::]/64', allow_network=True))

    def test_ipv6_range(self):
        """Test for IPv6 range."""
        self.assertFalse(is_ipv6_address('-'))
        self.assertFalse(is_ipv6_address('127.0.0.1-127.0.0.2'))
        self.assertFalse(is_ipv6_address('1::/64-1::/64'))
        self.assertFalse(is_ipv6_address('::-::/-64'))
        self.assertFalse(is_ipv6_address('::2-::1'))
        self.assertTrue(is_ipv6_address('::-::'))
        self.assertTrue(is_ipv6_address('::1-::2'))


class TestIsIPAddress(unittest.TestCase):
    """Basic unit tests for test_ip_address()."""

    def test_ip_address(self):
        """Test address detection for allow_negative=True/False."""
        self.assertEqual(is_ip_address('-127.0.0.1', allow_negative=True), 4)
        self.assertEqual(is_ip_address('-::', allow_negative=True), 6)
        self.assertEqual(is_ip_address('-127.0.0.1', allow_negative=False), 0)
        self.assertEqual(is_ip_address('-::', allow_negative=False), 0)
